//------------------------------------------------------------------------
// Copyright (C) 1996:
// Scientific computing department
// Sandia National Laboratories
//------------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#include "OPT++_config.h"
#endif

#ifdef HAVE_STD
#include <cstring>
#else
#include <string.h>
#endif

#include "OptBCQNewton.h"
#include <float.h>
#include "cblas.h"
#include "ioformat.h"

using Teuchos::SerialDenseVector;
using Teuchos::SerialDenseMatrix;
using Teuchos::SerialSymDenseMatrix;

namespace OPTPP {

static char* class_name = "OptBCQNewton";

//------------------------------------------------------------------------
// BCQNewton functions
// initHessian
// checkConvg
// checkDeriv
// initOpt
// printStatus
// stepTolNorm
// updateH
// computeSearch
// updateConstraints
// reset
//------------------------------------------------------------------------
void OptBCQNewton::initHessian()
{ 
  NLP1* nlp = nlprob();
  int   i,n = nlp->getDim();
  Hessian.reshape(n);
  Hessian = 0.0;
  for (i=0; i<n; i++) Hessian(i,i) = 1.0;
  return;
}

int OptBCQNewton::checkConvg() // Check convergence
{
  NLP1* nlp = nlprob();
  SerialDenseVector<int,double> xc(nlp->getXc());
  int   i, n = nlp->getDim();

  // Test 1. step tolerance 
  double step_tol = tol.getStepTol();
  double snorm = stepTolNorm();
  //double xnorm =  Norm2(xc);
  double xnorm = sqrt(xc.dot(xc));
  double stol  = step_tol*max(1.0,xnorm);
  if (snorm  <= stol) {
    strcpy(mesg,"Algorithm converged - Norm of last step is less than step tolerance");
    *optout << "checkConvg: snorm = " << e(snorm,12,4) 
      << "  stol = " << e(stol,12,4) << "\n";
    return 1;
  }
  
  // Test 2. function tolerance
  double ftol = tol.getFTol();
  double fvalue = nlp->getF();
  double rftol = ftol*max(1.0,fabs(fvalue));
  double deltaf = fprev - fvalue;
  if (deltaf <= rftol) {
    strcpy(mesg,"Algorithm converged - Difference in successive fcn values less than tolerance");
    *optout << "checkConvg: deltaf = " << e(deltaf,12,4) 
         << "  ftol = " << e(ftol,12,4) << "\n";
    return 2;
  }

  // Test 3. gradient tolerance 
  SerialDenseVector<int,double> grad(nlp->getGrad());
  double gtol = tol.getGTol();
  double rgtol = gtol*max(1.0,fabs(fvalue));
  for (i=0; i<n; i++) if (work_set(i) == true) grad(i) = 0.0;
  //double gnorm = Norm2(grad);
  double gnorm = sqrt(grad.dot(grad));
  if (gnorm <= rgtol) {
    strcpy(mesg,"Algorithm converged - Norm of gradient is less than gradient tolerance");
    *optout << "checkConvg: gnorm = " << e(gnorm,12,4) 
      << "  gtol = " << e(rgtol, 12,4) << "\n";
    return 3;
  }
  
  // Test 4. absolute gradient tolerance 
  if (gnorm <= gtol) {
    strcpy(mesg,"Algorithm converged - Norm of gradient is less than gradient tolerance");
    *optout << "checkConvg: gnorm = " << e(gnorm,12,4) 
      << "  gtol = " << e(gtol, 12,4) << "\n";
    return 4;
  }
  
  // Nothing to report 
  return 0;
}

int OptBCQNewton::checkDeriv() // Check the analytic gradient with FD gradient
{ return checkAnalyticFDGrad(); }

void OptBCQNewton::initOpt()
{ 
  OptBCNewtonLike::initOpt();
  updateConstraints(0);
  return;
}

void OptBCQNewton::printStatus(char *s) // set Message
{
  NLP1* nlp = nlprob();

  if (debug_) *optout << class_name << "::printStatus: \n";

  *optout << "\n\n=========  " << s << "  ===========\n\n";
  *optout << "Optimization method       = " << method << "\n";
  *optout << "Dimension of the problem  = " << nlp->getDim()   << "\n";
  *optout << "No. of bound constraints  = " << nlp->getDim()   << "\n";
  *optout << "Return code               = " << ret_code << " ("
       << mesg << ")\n";
  *optout << "No. iterations taken      = " << iter_taken  << "\n";
  *optout << "No. function evaluations  = " << nlp->getFevals() << "\n";
  *optout << "No. gradient evaluations  = " << nlp->getGevals() << "\n";

  if (debug_) {
    *optout << "Hessian \n";
    Print(Hessian);
  }

  tol.printTol(optout);

  nlp->fPrintState(optout, s);
}

double OptBCQNewton::stepTolNorm() const
{
  NLP1* nlp = nlprob();
  // SerialDenseVector<int,double> step(sx.AsDiagonal()*(nlp->getXc() - xprev));
  // return Norm2(step);
  SerialDenseVector<int,double> tmp(nlp->getXc().length());
  tmp = nlp->getXc();
  tmp -= xprev;
  SerialDenseVector<int,double> step(tmp.length());
  for(int i=0; i<tmp.length(); i++)
    {step(i) = tmp(i)*sx(i);}
  return sqrt(step.dot(step));
}

SerialSymDenseMatrix<int,double> OptBCQNewton::updateH(SerialSymDenseMatrix<int,double>& Hk, int k) 
{
  double mcheps = DBL_EPSILON;
  double sqrteps = sqrt(mcheps);

  NLP1* nlp = nlprob();
  int i, nr = nlp->getDim();
  SerialDenseVector<int,double> grad(nr), xc(nlp->getXc().length());
  xc     = nlp->getXc();
  grad   = nlp->getGrad();

  SerialDenseVector<int,double> D(nr);

  // BFGS formula
  if (k == 0) { 
    Hessian = 0.0;
    double typx, xmax, gnorm;

    // Initialize xmax, typx and D to default values
    xmax   = -1.e30; typx   =  1.0; D      =  1.0;

    gnorm = sqrt(grad.dot(grad));

    for (i=0; i < nr; i++) xmax = max(xmax,xc(i));
    if(xmax != 0.0) typx = xmax;
    if(gnorm!= 0.0) D    = gnorm/typx;
    if (debug_) {
      *optout << "updateH: gnorm0 = " << gnorm
	<< "typx = " << typx << "\n";
    }
    for (i=0; i < nr; i++) Hessian(i,i) = D(i);
    return Hessian;
  }
  
  // update the portion of H corresponding to the free variable list only

  SerialDenseVector<int,double> yk(nr), sk(nr), Bsk(nr);
  SerialDenseMatrix<int,double> Htmp(nr,nr);
  
  yk = grad;
  yk  -= gprev;
  sk = xc;
  sk -= xprev;
  for (i=0; i<nr; i++) {
    if (work_set(i) == true) {
      yk(i) = sk(i) = 0.0;
      //*optout << "fixed variable = " << i << "\n";
    }
  }
  
  double gts = gprev.dot(sk);
  double yts = yk.dot(sk);
  
  double snorm = sqrt(sk.dot(sk));
  double ynorm = sqrt(yk.dot(yk));
  
  if (debug_) {
    *optout << "updateH: gts   = " << gts 
         << "  yts = " << yts << "\n";
    *optout << "updateH: snorm = " << snorm 
         << "  ynorm = " << ynorm << "\n";
  }

  if (yts <= sqrteps*snorm*ynorm) {
    if (debug_) {
      *optout << "updateH: <y,s> = " << e(yts,12,4) << " is too small\n";
      *optout << "updateH: The BFGS update is skipped\n";
    }
    Hessian = Hk; return Hk;
  }
  
  SerialDenseVector<int,double> res(nr),tmp3(Hk.numRows());
  //res = yk - Hk*sk;
  tmp3.multiply(Teuchos::LEFT_SIDE, 1.0, Hk, sk, 0.0);
  res = yk;
  res -= tmp3;
  for (i=0; i<nr; i++) if (work_set(i) == true) res(i) = 0.0;
  if (res.normInf() <= sqrteps) {
    if (debug_) {
      *optout << "updateH: <y,s> = " << e(yts,12,4) << " is too small\n";
      *optout << "updateH: The BFGS update is skipped\n";
    }
    Hessian = Hk; return Hk;
  }
  
  //Bsk = Hk*sk;
  Bsk.multiply(Teuchos::LEFT_SIDE, 1.0, Hk, sk, 0.0);
  for (i=0; i<nr; i++) if (work_set(i) == true) Bsk(i) = 0.0;
  double sBs = sk.dot(Bsk);
  double etol = 1.e-8;

  if (sBs <= etol*snorm*snorm) {
    if (debug_) {
      *optout << "updateH: <y,s> = " << e(yts,12,4) << " is too small\n";
      *optout << "updateH: The BFGS update is skipped\n";
    }
    //D = sx.AsDiagonal()*sx.AsDiagonal();
    Hk = 0;
    for (i=0; i < nr; i++) Hk(i,i) = sx(i)*sx(i);
    Hessian = Hk; return Hk;
  }
  
  //Htmp = - (Bsk * Bsk.t()) / sBs;
 Htmp.multiply(Teuchos::NO_TRANS, Teuchos::TRANS, -1/sBs, Bsk, Bsk, 0.0);

 //Htmp = Htmp + (yk * yk.t()) / yts;
 // Htmp = Hk + Htmp;
 // Hk << Htmp;
 // Htmp.Release();
 SerialDenseMatrix<int,double> tmp4(yk.length(),yk.length());
  for(int i=0; i<yk.length();i++)
    for(int j=0; j<yk.length();j++)
      {tmp4(i,j) = yk(i)*yk(j)/yts;}
 Htmp += tmp4;
 for(int i=0; i<Htmp.numRows(); i++)
   for(int j=0; j<=i; j++)
     {Hk(i,j) = Hk(i,j) + Htmp(i,j);}
 
  SerialDenseVector<int,double> Bgk(nr), ggrad(nr);
  //Bgk = Hk*grad;
  Bgk.multiply(Teuchos::LEFT_SIDE, 1.0, Hk, grad, 0.0);
  ggrad = grad;
  for (i=0; i<nr; i++) if (work_set(i) == true) ggrad(i) = 0.0;
  double gBg = ggrad.dot(Bgk);
  double gg  = ggrad.dot(ggrad);
  double ckp1= gBg/gg;
  if (debug_) {
    *optout << "\nupdateH: after update, k = " << k << "\n";
    *optout << "updateH: sBs  = " << sBs << "\n";
    *optout << "updateH: ckp1 = " << ckp1 << "\n";
  }
  Hessian = Hk;
  return Hk;
}

//---------------------------------------------------------------------------- 
// Compute the maximum step allowed along the search direction sk
// before we hit a constraint
//--------------------------------------------------------------------------- 
double OptBCQNewton::computeMaxStep(SerialDenseVector<int,double> &sk)
{
  NLP1* nlp = nlprob();
  int i, n = nlp->getDim();
  double gamma=FLT_MAX, delta;
  SerialDenseVector<int,double> lower(nlp->getConstraints()->getLower().length());
  lower = nlp->getConstraints()->getLower();
  SerialDenseVector<int,double> upper(nlp->getConstraints()->getUpper());
  upper = nlp->getConstraints()->getUpper();
  SerialDenseVector<int,double> xc(nlp->getXc().length());
  xc    = nlp->getXc();

  double snorm = sqrt(sk.dot(sk));
  double feas_tol = 1.e-3;

  for (i=0; i<n; i++) {
    if (work_set(i) == false) {
      if      (sk(i) > 0.0e0) {
        delta = (upper(i)-xc(i)) / sk(i);
        if (delta <= feas_tol) {
	  if (debug_)
	    *optout << "Hit an upper constraint for variable " << i << "\n";
	}
      }
      else if (sk(i) < 0.0e0) {
        delta = (lower(i)-xc(i)) / sk(i);
        if (delta <= feas_tol) {
	  if (debug_)
	    *optout << "Hit a  lower constraint for variable " << i << "\n";
	}
      }
      gamma = min(gamma,delta);
    }
  }
  if (debug_)
    *optout << "computeMaxStep: maximum step allowed = " << gamma*snorm << "\n";
  return gamma*snorm;
}

SerialDenseVector<int,double> OptBCQNewton::computeSearch(SerialSymDenseMatrix<int,double> &H)
{
  NLP1*	nlp = nlprob();
  int   i, j, ncnt=0, *index_array, n = nlp->getDim();

  SerialDenseVector<int,double>          gg(n), sk2(n), sk(n);
  SerialSymDenseMatrix<int,double>       H1;
  SerialDenseMatrix<int,double> L;

  // set up index_array to count the number of free variables

  index_array = new int[n+1];
  for (i=1; i<=n; i++) index_array[i] = 0;
  for (i=1; i<=n; i++) 
    if (work_set(i-1) == false) index_array[i] = ++ncnt;
  if (ncnt != (n-nactive)) {
    *optout << "Number of fixed and free variables do not correspond. \n";
    exit(-1);
  }

  // Form the projected Hessian

  H1.reshape(ncnt);
  for (i=1; i<=n; i++) {
     for (j=1; j<=n; j++) 
       if (index_array[i] != 0 && index_array[j] != 0)  
	  H1(index_array[i]-1,index_array[j]-1) = H(i-1,j-1);
  }

  // Form the projected gradient

  gg.resize(ncnt);
  for (i=1; i<=n; i++) 
    if (index_array[i] != 0) gg(index_array[i]-1) = gprev(i-1); 

  // Solve (H1 * sk2 = - gg) for projected search direction sk2 

  L.reshape(ncnt,ncnt);
  sk2.resize(ncnt);
  if (ncnt == 1) sk2(0) = - gg(0) / H1(0,0);
  else if (ncnt > 1) {
    L   = MCholesky(H1);
    //sk2 = -(L.t().i()*(L.i()*gg));
    Teuchos::LAPACK<int,double> lapack;
    sk2 = gg;
    sk2 *= -1;
    int INFO;
    lapack.TRTRS('L','N','N',ncnt,1,L.values(),ncnt, sk2.values(),ncnt, &INFO);
    lapack.TRTRS('L','T','N',ncnt,1,L.values(),ncnt,sk2.values(),ncnt,&INFO);
    //return sk2;

    // Teuchos::SerialSpdDenseSolver<int,double> My_Solver;
    // int info = 0;
 
    // My_Solver.setMatrix(Teuchos::rcp(&H1,false));
    // My_Solver.setVectors(Teuchos::rcp(&sk2, false), Teuchos::rcp(&gg,false));
    // My_Solver.equilibrateMatrix();
    //My_Solver.equilibrateRHS();
    // info = My_Solver.solve();
    // if(info != 0)
    // {return sk2;}

  }


  // Form search direction sk from from projected search direction sk2 

  for (i=0; i<n; i++) sk(i) = 0.0;
  for (i=1; i<=n; i++) 
    if (index_array[i] != 0) sk(i-1) = sk2(index_array[i]-1);

  // Sanitation and return

  delete [] index_array; 
  return sk;
}

int OptBCQNewton::updateConstraints(int step_type)
{
  NLP1*    	nlp = nlprob();
  int          	n = nlp->getDim(), ret_flag=0;
  int          	i, j, j2, k, *new_active, actcnt=0, notnew;
  double       	reduced_grad_norm, feas_tol=1.0e-12;
  SerialDenseVector<int,double> 	lower(n), upper(n), xc(n), gg(n);

  // initialization

  lower      = nlp->getConstraints()->getLower();
  upper      = nlp->getConstraints()->getUpper();
  xc         = nlp->getXc();
  new_active = new int[n];

  // Add variables to the working set

  for (i=0; i<n; i++) {
    if ((fabs(upper(i)-xc(i))<feas_tol) || (fabs(lower(i)-xc(i))<feas_tol)) {
      if (work_set(i) == false) {
        new_active[actcnt++] = i; work_set(i) = true; nactive++;
        *optout << "OptBCQNewton : variable added to working set : " << i << "\n";
      }
    } 
  }

  // Delete variables from the active set 
  // First compute the norm of the reduced gradient

  int    jdel=-1;
  double ggdel=0;

  gg = nlp->getGrad();
  for (i=0; i<n; i++) if(work_set(i) == true) gg(i) = 0.0;
  reduced_grad_norm = sqrt(gg.dot(gg));
  if (m_nconvgd > 0 || step_type < 0) {
    gg = nlp->getGrad();
    ret_flag = -1;
    *optout << "OptBCQNewton : reduced_grad_norm = " << reduced_grad_norm << "\n";
    for (i=0; i<n; i++) {
      notnew = true;
      for (j=0; j<actcnt; j++) if (new_active[j] == i) notnew = false;
      if (work_set(i) == true && notnew) 
        if (((fabs(upper(i)-xc(i))<feas_tol) && gg(i)>feas_tol) ||
            ((fabs(lower(i)-xc(i))<feas_tol) && gg(i)<(-feas_tol))) {
	  if (fabs(gg(i)) > ggdel) {jdel = i; ggdel = fabs(gg(i)); }
	}
    }
    if (jdel != -1) {
      work_set(jdel) = false; nactive--;
      *optout << "OptBCQNewton : variable deleted from working set : " << jdel << "\n";
      ret_flag = 1;
    }
  }
  if (nactive > 0) *optout << "OptbCNewton: Current working set  \n";
  k = 1;
  for (i=1; i<=nactive; i+=10) {
    *optout << " ----- variables : ";
    j2 = min(i*10,nactive);
    for (j=(i-1)*10+1; j<=j2; j++) {
      while (work_set(k-1) == false) k++;
      *optout << d(k-1,6); k++;
    }
    *optout << "\n ";
  }
  return ret_flag;
}

void OptBCQNewton::reset()
{
   NLP1* nlp = nlprob();
   int   n   = nlp->getDim();
   nlp->reset();
   OptimizeClass::defaultReset(n);
   nactive  = 0;
   work_set = false;
}

} // namespace OPTPP
