// $Id: HOPSPACK_EvaluatorDefault.hpp 149 2009-11-12 02:40:41Z tplante $
// $URL: https://software.sandia.gov/svn/public/hopspack/trunk/src/src-evaluator/HOPSPACK_EvaluatorDefault.hpp $

//@HEADER
// ************************************************************************
// 
//         HOPSPACK: Hybrid Optimization Parallel Search Package
//                 Copyright 2009 Sandia Corporation
// 
// Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
// the U.S. Government retains certain rights in this software.
// 
// This file is part of HOPSPACK.
//
// HOPSPACK is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version.
//  
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//  
// You should have received a copy of the GNU Lesser General Public
// License along with this library.  If not, see http://www.gnu.org/licenses/.
// 
// Questions? Contact Tammy Kolda (tgkolda@sandia.gov)
//                 or Todd Plantenga (tplante@sandia.gov) 
// 
// ************************************************************************
//@HEADER

/*!
  @file HOPSPACK_EvaluatorDefault.hpp
  @brief Declaration for HOPSPACK::EvaluatorDefault, subclass of Evaluator.
*/

#ifndef HOPSPACK_EVALUATORDEFAULT_HPP
#define HOPSPACK_EVALUATORDEFAULT_HPP

#include "HOPSPACK_common.hpp"
#include "HOPSPACK_Evaluator.hpp"
#include "HOPSPACK_ParameterList.hpp"
#include "HOPSPACK_Vector.hpp"

namespace HOPSPACK
{


//----------------------------------------------------------------------
//! Interface for default evaluation of functions and nonlinear constraints.
/*!
 *  This is the default implementation shipped with HOPSPACK.
 *  The default is compatible with serial, MPI, and multithreaded versions
 *  of HOPSPACK.
 *
 *  The value of configuration parameter "Evaluator Type" determines at
 *  runtime how the evaluation is made:
 *    "System Call" - execute an application program using a system call
 *    "AMPL"        - evaluate an AMPL .nl file using the AMPL Solver Library
 */
//----------------------------------------------------------------------
class EvaluatorDefault : public Evaluator
{
  public:

    //! Parameterized factory method for constructing an instance.
    /*!
     *  @param[in] cEvalParams  Parameters in the "Evaluator" sublist.
     *                          Parameter value "Evaluator Type" determines
     *                          the particular implementation.
     *  @return                 New Evaluator instance if successful, else NULL.
     *                          Client must delete the instance when finished.
     */
    static EvaluatorDefault *  newInstance (const ParameterList &  cEvalParams);

    //! Destructor.
    ~EvaluatorDefault (void);


    //! Evaluate the objective function(s) at a point x.
    /*!
     *  @param[in] nTag   Contains a unique tag for the evaluation which can be
     *                    used to name files, etc.
     *  @param[in] cX     The point at which to evaluate the function(s).
     *  @param[out] cFns  On output, contains a vector of objective function
     *                    values computed at X.  Multiple objectives are allowed.
     *                    If an evaluation failed, return an empty vector or set
     *                    individual elements of the vector to HOPSPACK::dne().
     *  @param[out] sMsg  On output, contains a message about the evaluation;
     *                    typically the word "Success" or an error message.
     */
    virtual void  evalF (const int       nTag,
                         const Vector &  cX,
                               Vector &  cFns,
                               string &  sMsg) = 0;

    //! Evaluate the objective functions and nonlinear constraints at a point x.
    /*!
     *  @param[in] nTag     Contains a unique tag for the evaluation which can be
     *                      used to name files, etc.
     *  @param[in] cX       The point at which to evaluate the function(s).
     *  @param[out] cFns    On output, contains a vector of objective function
     *                      values computed at X.  Multiple objectives are
     *                      allowed.  If an evaluation failed, return an empty
     *                      vector or set individual function elements of the
     *                      vector to HOPSPACK::dne().
     *  @param[out] cEqs    On output, contains a vector of nonlinear equality
     *                      constraint function values computed at X.  If an
     *                      evaluation failed, return an empty vector or set
     *                      individual elements of the vector to HOPSPACK::dne().
     *  @param[out] cIneqs  On output, contains a vector of nonlinear inequality
     *                      constraint function values computed at X.  If an
     *                      evaluation failed, return an empty vector or set
     *                      individual elements of the vector to HOPSPACK::dne().
     *  @param[out] sMsg    On output, contains a message about the evaluation;
     *                      typically the word "Success" or an error message.
     */
    virtual void  evalFC (const int       nTag,
                          const Vector &  cX,
                                Vector &  cFns,
                                Vector &  cEqs,
                                Vector &  cIneqs,
                                string &  sMsg) = 0;

    //! Print debug information about the Evaluator instance.
    virtual void  printDebugInfo (void) const = 0;

  protected:

    //! Default constructor is hidden, forcing clients to use the factory method.
    EvaluatorDefault (void);


    static const char  sEVAL_PARAM_ERROR[];
};

}          //-- namespace HOPSPACK

#endif     //-- HOPSPACK_EVALUATORDEFAULT_HPP
