/*  _______________________________________________________________________

    PECOS: Parallel Environment for Creation Of Stochastics
    Copyright (c) 2011, Sandia National Laboratories.
    This software is distributed under the GNU Lesser General Public License.
    For more information, see the README file in the top Pecos directory.
    _______________________________________________________________________ */

//- Class:	 TensorProductDriver
//- Description: 
//- Owner:       Mike Eldred
//- Revised by:  
//- Version:

#ifndef INTEGRATION_DRIVER_HPP
#define INTEGRATION_DRIVER_HPP

#include "pecos_data_types.hpp"
#include "BasisPolynomial.hpp"

namespace Pecos {

class DistributionParams;
class BasisConfigOptions;


/// base class for generating N-dimensional grids for numerical evaluation
/// of expectation integrals over independent standard random variables.

/** This class enables Dakota::NonD{Quadrature,Cubature,SparseGrid}. */

class IntegrationDriver
{
public:

  //
  //- Heading: Constructors, destructor, and operator=
  //

  /// default constructor
  IntegrationDriver();
  /// standard constructor for envelope
  IntegrationDriver(short driver_type);
  /// copy constructor
  IntegrationDriver(const IntegrationDriver& driver);

  /// destructor
  virtual ~IntegrationDriver();

  /// assignment operator
  IntegrationDriver operator=(const IntegrationDriver& driver);

  //
  //- Heading: Virtual functions
  //

  /// update polynomialBasis with data from dist_params
  virtual void initialize_grid_parameters(const ShortArray& u_types,
					  const DistributionParams& dp);

  /// compute scaled variable and weight sets for the TPQ grid
  virtual void compute_grid();
  /// compute number of collocation points
  virtual int grid_size();

  //
  //- Heading: Member functions
  //

  /// return polynomialBasis
  const std::vector<BasisPolynomial>& polynomial_basis() const;

  /// return type1WeightSets
  const RealVector& type1_weight_sets() const;
  /// return type2WeightSets
  const RealMatrix& type2_weight_sets() const;

  // append to end of type1WeightSets
  //void append_type1_weight_sets(const RealVector& t1_wts);
  // append to end of type2WeightSets
  //void append_type2_weight_sets(const RealMatrix& t2_wts);

  /// return collocPts1D
  const Real3DArray& collocation_points_array()  const;
  /// return type1CollocWts1D
  const Real3DArray& type1_collocation_weights_array() const;
  /// return type2CollocWts1D
  const Real3DArray& type2_collocation_weights_array() const;

  /// return collocRules
  const ShortArray& collocation_rules() const;

  /// return orderGenzKeister
  const UShortArray& genz_keister_order()     const;
  /// return precGenzKeister
  const UShortArray& genz_keister_precision() const;

  /// returns approxRep for access to derived class member functions
  /// that are not mapped to the base level
  IntegrationDriver* driver_rep() const;

protected:

  //
  //- Heading: Constructors
  //

  /// constructor initializes the base class part of letter classes
  /// (BaseConstructor overloading avoids infinite recursion in the
  /// derived class constructors - Coplien, p. 139)
  IntegrationDriver(BaseConstructor);

  //
  //- Heading: Member functions
  //

  /// set int_rules and growth_rules from u_types and mode booleans
  void initialize_rules(const ShortArray& u_types,
			const Pecos::BasisConfigOptions& bc_options);
  /// set int_rules and growth_rules from poly_basis and growth_rate
  void initialize_rules(const std::vector<BasisPolynomial>& poly_basis);

  /// compute variable and weight sets for a tensor-product grid
  void compute_tensor_grid(const UShortArray& quad_order,
			   const UShortArray& lev_index,
			   RealMatrix& variable_sets,
			   RealVector& t1_weight_sets,
			   RealMatrix& t2_weight_sets,
			   UShort2DArray& colloc_key);

  /// update collocPts1D[lev_index][i] and type{1,2}CollocWts1D[lev_index][i]
  void assign_1d_collocation_points_weights(size_t i, unsigned short quad_order,
					    unsigned short lev_index);

  //
  //- Heading: Data
  //

  /// number of variables in the tensor-product grid
  size_t numVars;

  /// enumeration codes for integration rule options.  Manages internal
  /// mode switches for 1D polynomial types: e.g., GAUSS_LEGENDRE or
  /// GAUSS_PATTERSON for Legendre, CLENSHAW_CURTIS or FEJER2 for
  /// Chebyshev, GAUSS_HERMITE or GENZ_KEISTER for Hermite.
  ShortArray collocRules;

  /// array of one-dimensional orthogonal polynomials used in
  /// computing Gaussian quadrature points and weights
  std::vector<BasisPolynomial> polynomialBasis;

  /// the set of type1 weights (for integration of value interpolants)
  /// associated with each point in the {TPQ,SSG,Cub} grid
  RealVector type1WeightSets;
  /// the set of type2 weights (for integration of gradient interpolants)
  /// for each derivative component and for each point in the {TPQ,SSG} grid
  RealMatrix type2WeightSets;

  /// num_levels_per_var x numVars sets of 1D collocation points
  Real3DArray collocPts1D;
  /// num_levels_per_var x numVars sets of 1D type1 collocation weights
  Real3DArray type1CollocWts1D;
  /// num_levels_per_var x numVars sets of 1D type2 collocation weights
  Real3DArray type2CollocWts1D;

  /// flag indicating usage of compute1DType2Weights to define type2WeightSets
  bool computeType2Weights;

  /// lookup for set of 1-D Genz-Keister quadrature orders
  static UShortArray orderGenzKeister;
  /// lookup for set of 1-D Genz-Keister integrand precisions
  static UShortArray precGenzKeister;

private:

  //
  //- Heading: Member functions
  //

  /// update collocPts1D and type{1,2}CollocWts1D
  void update_1d_collocation_points_weights(const UShortArray& quad_order,
					    const UShortArray& lev_index);

  /// Used only by the standard envelope constructor to initialize
  /// basisApproxRep to the appropriate derived type.
  IntegrationDriver* get_driver(short driver_type);

  //
  //- Heading: Data members
  //

  /// pointer to the letter (initialized only for the envelope)
  IntegrationDriver* driverRep;
  /// number of objects sharing driverRep
  int referenceCount;
};


inline const std::vector<BasisPolynomial>& 
IntegrationDriver::polynomial_basis() const
{ return (driverRep) ? driverRep->polynomialBasis : polynomialBasis; }


inline const RealVector& IntegrationDriver::type1_weight_sets() const
{ return (driverRep) ? driverRep->type1WeightSets : type1WeightSets; }


inline const RealMatrix& IntegrationDriver::type2_weight_sets() const
{ return (driverRep) ? driverRep->type2WeightSets : type2WeightSets; }


/*
inline void IntegrationDriver::
append_type1_weight_sets(const RealVector& t1_wts)
{
  if (driverRep)
    driverRep->append_type1_weight_sets(t1_wts);
  else {
    size_t i, num_curr_t1_wts = type1WeightSets.length(),
      num_new_t1_wts = t1_wts.length(),
      num_total_t1_wts = num_curr_t1_wts + num_new_t1_wts;
    type1WeightSets.resize(num_total_t1_wts);
    for (i=0; i<num_new_t1_wts; ++i)
      type1WeightSets[num_curr_t1_wts+i] = t1_wts[i];
  }
}


inline void IntegrationDriver::
append_type2_weight_sets(const RealMatrix& t2_wts)
{
  if (driverRep)
    driverRep->append_type2_weight_sets(t2_wts);
  else {
    size_t i, j, num_curr_t2_wts = type2WeightSets.numCols(),
      num_new_t2_wts = t2_wts.numCols(),
      num_total_t2_wts = num_curr_t2_wts + num_new_t2_wts;
    type2WeightSets.reshape(numVars, num_total_t2_wts);
    for (i=0; i<num_new_t2_wts; ++i) {
      Real*      curr_t2_i = type2WeightSets[num_curr_t2_wts+i];
      const Real* new_t2_i = t2_wts[i];
      for (j=0; j<numVars; ++j)
	curr_t2_i[j] = new_t2_i[j];
    }
  }
}
*/


inline const Real3DArray& IntegrationDriver::collocation_points_array() const
{ return collocPts1D; }


inline const Real3DArray& IntegrationDriver::
type1_collocation_weights_array() const
{ return type1CollocWts1D; }


inline const Real3DArray& IntegrationDriver::
type2_collocation_weights_array() const
{ return type2CollocWts1D; }


inline const ShortArray& IntegrationDriver::collocation_rules() const
{ return collocRules; }


inline const UShortArray& IntegrationDriver::genz_keister_order() const
{ return (driverRep) ? driverRep->orderGenzKeister : orderGenzKeister; }


inline const UShortArray& IntegrationDriver::genz_keister_precision() const
{ return (driverRep) ? driverRep->precGenzKeister : precGenzKeister; }


inline IntegrationDriver* IntegrationDriver::driver_rep() const
{ return driverRep; }

} // namespace Pecos

#endif
