/*  _______________________________________________________________________

    DAKOTA: Design Analysis Kit for Optimization and Terascale Applications
    Copyright (c) 2010, Sandia National Laboratories.
    This software is distributed under the GNU Lesser General Public License.
    For more information, see the README file in the top Dakota directory.
    _______________________________________________________________________ */

//- Class:       MixedVariables
//- Description: Container class of variables, employing a "mixed"
//-              continuous/discrete approach to variables usage.
//- Owner:       Mike Eldred
//- Version: $Id: MixedVariables.H 7024 2010-10-16 01:24:42Z mseldre $

#ifndef MIXED_VARS_H
#define MIXED_VARS_H

#include "DakotaVariables.H"
#include "DataVariables.H"


namespace Dakota {

/// Derived class within the Variables hierarchy which employs
/// the default data view (no variable or domain type array merging).

/** Derived variables classes take different views of the design,
    uncertain, and state variable types and the continuous and
    discrete domain types.  The MixedVariables derived class
    separates the design, uncertain, and state variable types as well
    as the continuous and discrete domain types.  The result is
    separate arrays for continuous design, discrete design, uncertain,
    continuous state, and discrete state variables.  This is the
    default approach, so all iterators and strategies not specifically
    utilizing the All or Merged views use this approach (see
    Variables::get_variables(problem_db)). */

class MixedVariables: public Variables
{
public:

  //
  //- Heading: Constructors and destructor
  //

  /// standard constructor
  MixedVariables(const ProblemDescDB& problem_db,
		 const std::pair<short,short>& view);
  /// lightweight constructor
  MixedVariables(const SharedVariablesData& svd);
  /// destructor
  ~MixedVariables();

protected:

  //
  //- Heading: Virtual function redefinitions
  //

  void read(std::istream& s);
  void write(std::ostream& s) const;
  void write_aprepro(std::ostream& s) const;
  void read_tabular(std::istream& s);
  void write_tabular(std::ostream& s) const;

  void reshape(const SizetArray& vc_totals);

  void build_active_views();
  void build_inactive_views();

private:

  //
  //- Heading: Private member functions
  //
};


inline MixedVariables::MixedVariables(const SharedVariablesData& svd):
  Variables(BaseConstructor(), svd)
{ reshape(svd.components_totals()); }


inline MixedVariables::~MixedVariables()
{ }

} // namespace Dakota

#endif
